/* This is source for a quickie hack to access
 * ACPI power management stuff.
 *
 * It is based around my "generic pci base" framework, from
 *    http://www.bolthole.com/solaris/drivers/
 *
 *
 * Read the README file for important information
 *    source version @(#) acpi.c 1.9@(#)
 * 
 * You can use this source for any purpose you like [without any warantee, etc]
 * It would be nice if you give me some credit if you use it, but you dont
 * have to.
 *
 *
 *  DISCLAIMER: This code may not be 100% clean, kernel-wise.
 *  But it is as clean as I know how to make it, with lots of
 *  hopefully helpful comments.
 *    Philip Brown
 */

#include <stdarg.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>

#include <sys/ksynch.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>



/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)

/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"acpi"	
#define DEVNODE_TYPE	DDI_PSEUDO /* see sunddi.h */

/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [acpi_cb_ops, acpi_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	acpi_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	acpi_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	acpi_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	acpi_print(dev_t, char*);

static	int	acpi_read(dev_t dev, struct uio* uiop, cred_t * credp);

static	int	acpi_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	acpi_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	acpi_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	acpi_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;

/* "char/block operations" */
static struct cb_ops	acpi_cb_ops = {
	acpi_open,      /* open() */
	acpi_close,     /* close() */
	nodev,          /* strategy() */
	acpi_print,     /* print routine */
	nodev,          /* no dump routine */
	acpi_read,      /* read() */
	acpi_write,     /* write() */
	acpi_ioctl,     /* generic ioctl */
	nodev,          /* no devmap routine */
	nodev,          /* no mmap routine */
	nodev,          /* no segmap routine */
	nochpoll,       /* no chpoll routine */
	ddi_prop_op,
	0,              /* not a STREAMS driver	*/
	D_NEW | D_MP,   /* safe for multi-thread/multi-processor */
	CB_REV,         /* cb_ops version? */
	nodev,          /* cb_aread() */
	nodev,          /* cb_awrite() */
};

/* "device operations". */
static struct dev_ops acpi_ops = {
	DEVO_REV,            /* devo_rev */
	0,                   /* devo_refcnt */
	acpi_getinfo,        /* devo_getinfo */
	nulldev,             /* devo_identify */
	nulldev,             /* devo_probe */
	acpi_attach,         /* devo_attach */
	acpi_detach,         /* devo_detach */
	nodev,               /* devo_reset */
	&acpi_cb_ops,        /* devo_cb_ops */
	(struct bus_ops *)0, /* devo_bus_ops */
	NULL,                /* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"acpi driver from bolthole.com",
	&acpi_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


/* This describes things like byteorder internally for the pci card */
static struct ddi_device_acc_attr dev_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};


#define MAX_INSTANCES 1 /* max of 1 instance allowed */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   acpi_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * acpi_glob_soft_handle;

/* This gets filled out on a per-instance basis by acpi_attach */
struct acpi_softstate {
	dev_info_t *dip; /* you'll probably want this saved! */
	kmutex_t instance_mutex;
	kmutex_t intr_mutex;
	ddi_iblock_cookie_t intr_cookie;

	kcondvar_t wait_cv;
	kmutex_t wait_mutex;
	int done_mutexes;

	int enable_powerbutton;

	int acpi_version;
	ddi_acc_handle_t acpi_regs_handle; /*opaque kernel stuff to map regs*/
	void *acpi_regs; /* actual pointer to mapped registerset*/
};



/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/

/********************************/
/* Register in PCI Config space */
/********************************/

#define PCICONF_ACPI_BASE	0x40
#define PCICONF_ACPI_CTL	0x44
  #define ACPI_CTL_ACPI_EN      (1 <<4 )
#define PCICONF_GEN_CTL		0xd0
  #define GEN_CTL_APIC_EN       (1 << 8)



/* These numbers are offsets from the ACPI 'base' address */
#define PM1_STS 0x0	/* STATUS */
  #define PM1_STS_PRBTNOR     (1 <<11) /* "powerbutton was forced" flag */
  #define PM1_STS_PRBTN_STS   (1 <<8) /* "powerbutton was forced" flag */
#define PM1_EN  0x2	/* ENABLE ACPI features */
  #define PM1_EN_PWRBTN_EN    (1 <<8)
  #define PM1_EN_GBL_EN       (1 <<5)
#define PM1_CNT 0x4	/* CONTROL */
  #define PM1_CNT_SCI_EN       0x01    /* use SCI instead of SMI# */
  #define SLP_EN	      (1<<13)  /* sleep enable */
  #define SLP_TYPE_OFF    ((1<<12)|(1<<11)|(1<<10))
  #define SLP_TYPE_DSUSP  ((1<<12)|(1<<11)) /* disk suspend */
  /* ACPI version 1 has different bit values for some things. sigh. */
  #define V1_SLP_TYPE_OFF 0




void
acpi_debugprint(const char *format, ...)
{
	va_list args;
	va_start(args, format);

	if(debug_level==0){
		return;
	}

	/* DO NOT USE vcmn_err here!!! it will not be interrupt-safe.
	 * See the "CONTEXT" part of the manpage
	 */

	vprintf(format, args);
	va_end(args);
}


/*Like the acpi_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See acpi_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	/*acpi_debugprint("acpi: _init called\n");*/

	/*
	 * Initialize global mutex before mod_install'ing driver.
	 * If mod_install() fails, must clean up mutex initialization
	 */
	mutex_init(&acpi_global_mutex, "acpi: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct acpi_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&acpi_glob_soft_handle,
				sizeof(struct acpi_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&acpi_global_mutex);
		ddi_soft_state_fini(&acpi_glob_soft_handle);
		cmn_err(CE_NOTE, "acpi: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	/*cmn_err(CE_NOTE, "acpi: _init returning SUCCESS, supposedly\n");*/
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	/*cmn_err(CE_NOTE, "acpi: _info called\n");*/
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	/*cmn_err(CE_NOTE, "acpi: _fini called\n");*/
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&acpi_glob_soft_handle);
		mutex_destroy(&acpi_global_mutex);
	}
	return (ret);
}


/**********************************************************************/

/* enable "APIC" stuff. Hmm */
static void enable_apic(struct acpi_softstate *softstate)
{
	uint32_t gen_ctl;
	ddi_acc_handle_t pcihandle;
	
	if(pci_config_setup(softstate->dip, &pcihandle) != DDI_SUCCESS){
		cmn_err(CE_WARN,"enable_apic: pci_config_setup FAILED??\n");
		return;
	}
	gen_ctl=pci_config_get32(pcihandle, 0xd0);
	if((gen_ctl & GEN_CTL_APIC_EN) ==0){
		printf("DEBUG: forcing APIC enable for gen_ctl register\n");
		gen_ctl |= GEN_CTL_APIC_EN;
		pci_config_put32(pcihandle, 0xd0, gen_ctl);
		printf("  .. done\n");
	}

	pci_config_teardown(&pcihandle);
}


/* This tweaks a register in the ACPI_BASE regset.
 *
 * SCI affects event notification for
 *   "events including PM1_STS bits 10,8,0 and the bits in GPE0_STS"
 */
static void enable_sci(struct acpi_softstate *softstate)
{
	uint32_t reg32;
	void * regaddr = (uint8_t*)softstate->acpi_regs;

	acpi_debugprint("enable_sci called\n");

	reg32=PM1_CNT_SCI_EN;

	ddi_put32(softstate->acpi_regs_handle, regaddr + PM1_CNT, reg32);
	
	reg32=ddi_get32(softstate->acpi_regs_handle, regaddr + PM1_CNT);
	acpi_debugprint("PM1_CNT value now shows as 0x%x\n",reg32);

}


/* return 1 on proper validation, 0 otherwise */
static int validate_regs(struct acpi_softstate *softstate)
{
	uint16_t vendorid;
	uint32_t acpibase;
	uint8_t acpi_ctl;
	uint32_t gen_ctl;
	ddi_acc_handle_t pcihandle;
	int instance, nregs=0;

	if(softstate->dip==NULL){
		cmn_err(CE_WARN,"validate_regs: internal err\n");
		return 0;
	}
	instance=ddi_get_instance(softstate->dip);

	if(pci_config_setup(softstate->dip, &pcihandle) != DDI_SUCCESS){
		cmn_err(CE_WARN,"validate_regs: pci_config_setup FAILED??\n");
		return 0;
	}

	vendorid=pci_config_get16(pcihandle, 0x0);
	acpibase=pci_config_get32(pcihandle, 0x40);
	acpi_ctl=pci_config_get8(pcihandle, 0x44);
	gen_ctl=pci_config_get32(pcihandle, 0xd0);


	pci_config_teardown(&pcihandle);

	if(! (acpibase&0x01)){
		cmn_err(CE_WARN,"acpi base register INVALID\n");
		return 0;
	}

	/* just throw away bit 0 first: that is always 1*/
	acpibase &=0xfffffffe;

	if(acpibase & 0xffff00ff){
		cmn_err(CE_WARN,"acpi base value looks wrong: 0x%x\n",
			acpibase);
		return 0;
	}


	/* now look at the hopefully good I/O space address */
	acpibase &=0x0000ff00;

	if(acpibase < 0x0400){ /* might actually want 0x800 here */
		cmn_err(CE_WARN,"acpi base value looks wrong: 0x%x\n",
			acpibase);
		return 0;
	}
		

	if(vendorid != 0x8086){
		cmn_err(CE_NOTE,"WARNING: vendor=0x%x, not 0x8086\n",
			vendorid);
	}

	ddi_dev_nregs(softstate->dip, &nregs);
	if(nregs==0){
		cmn_err(CE_WARN,"validate_regs: NO registers??\n");
		return 0;
	}

	if(nregs!=2){
		cmn_err(CE_WARN,"missing register specification\n");
		cmn_err(CE_WARN,"need the following in acpi.conf:\n");
		cmn_err(CE_WARN,
		  "  reg=[pci addrs...],0x0040f801,0x%x,0x%x,0x%x,0x%x;\n",
			0, acpibase, 0, 0x80/*size*/  );
		return 0;
	}

	acpi_debugprint("acpi_ctl from PCIconfig is 0x%x\n",acpi_ctl);
	acpi_debugprint("gen_ctl from PCIconfig is 0x%x\n",gen_ctl);

	acpi_debugprint("acpibase from PCIconfig is 0x%x\n",acpibase);

	if(softstate->acpi_regs==NULL){
		/* we've checked all we can right now. return good.*/
		return 1;
	}

	if(acpibase != (uint32_t)softstate->acpi_regs!=NULL){
		cmn_err(CE_WARN,
			"ERROR: ACPI map is not at acpibase 0x%x\n",acpibase);
		cmn_err(CE_WARN,
			"doing nasty tweak. Edit acpi.conf to fix this\n");
		softstate->acpi_regs = (void*)acpibase;
		acpi_debugprint("forced acpi_regs to 0x%p\n",
				softstate->acpi_regs);

	}

	return 1;
}

/**********************************************************************/

/*
 * interrupt handler for the device.
 * This wont ever get called until ddi_add_intr has been called.
 * Dont forget that you can only use intr_mutex in here, NOT
 * instance_mutex.
 * Additionally, you can ONLY USE FUNCTIONS that are
 * INTERRUPT CONTEXT SAFE.
 *  "user", or "kernel" context, just wont cut it.
 *
 * Note that cmn_err() is NOT interrupt-context safe.
 */
static uint_t
acpi_intr(caddr_t intr_arg)
{
	/* struct acpi_softstate *softstate;
	 * softstate = (struct acpi_softstate *)intr_arg;
	 */

	/* Put whatever you want here, IF you have interrupts */
	/* dont forget to call ddi_add_intr() to enable it! */

	return DDI_INTR_CLAIMED;
}

/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS
 */
static int acpi_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "acpi-print(%d): %s",instance,string);
	return 0;
}

static int acpi_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	acpi_debugprint("acpi: getinfo called\n");
	return DDI_FAILURE;
}





/* This routine can be called by acpi_attach (if errors occur)
 * and acpi_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void acpi_cleanup(struct acpi_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: acpi_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	if(softstate->intr_cookie != NULL){
		ddi_remove_intr(softstate->dip,0,softstate->intr_cookie);
	}
	if(softstate->acpi_regs_handle !=NULL){
		ddi_regs_map_free(&softstate->acpi_regs_handle);
	}

	/* This may not be strictly neccessary, but it is good coding
	 * practice
	 */

	if(softstate->done_mutexes){
		cv_destroy(&softstate->wait_cv);
		mutex_destroy(&softstate->wait_mutex);
		mutex_destroy(&softstate->instance_mutex);
	}

	ddi_soft_state_free(acpi_glob_soft_handle,instance);
	
}


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in acpi_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  acpi_detach()
 */
static int acpi_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	struct acpi_softstate *softstate;

	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 0);


	acpi_debugprint("acpi: acpi_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "acpi: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "acpi: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "acpi: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	acpi_debugprint("acpi: attach for instance %d\n", instance);

	if(instance!=0){
		cmn_err(CE_WARN,"ERR: only instance 0 is allowed, for acpi\n");
		return DDI_FAILURE;
	}

	/* Okay, since we only allow for 1 instance, we could theoretically
	 * hardcode this stuff. But that would be a bad example to set
	 * for driver writers. So we do the proper way here, reguardless
	 */
	/* Note: the instance check is a little redundant, because if
	 * we were over MAX_INSTANCES, these two calls would fail anyway.
	 * But it's nice to have an explicit warning message there.
	 */
	if (ddi_soft_state_zalloc(acpi_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "acpi: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(acpi_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "acpi: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(acpi_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	/* Default to acpi version 2. but allow user override to v1 */
	softstate->acpi_version=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "acpi-version", 2);

	/* check to see if we should enable ACPI control of powerbutton.
	 * This is not enabled by default, since there is a bug where
	 * it locks up laptops if the power cord is pulled.
	 */
	softstate->enable_powerbutton=ddi_prop_get_int(DDI_DEV_T_ANY,
	                            dip, DDI_PROP_DONTPASS,
                                    "enable-powerbutton", 0);

	acpi_debugprint("acpi_version=%d\n",softstate->acpi_version);

	softstate->dip = dip;


	if(!validate_regs(softstate)){
		cmn_err(CE_WARN, "acpi: attach for instance %d: failed check",
			instance);
		ddi_soft_state_free(acpi_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}


	/* reg "0",is the pci config register set (kinda) */
	if( ddi_regs_map_setup(dip, 1,
				(caddr_t *)&softstate->acpi_regs,
				0,0x80,
				&dev_attr, &softstate->acpi_regs_handle)
	   != DDI_SUCCESS)
	{
		cmn_err(CE_WARN,"ERR: regs_map_setup failed\n");
		ddi_soft_state_free(acpi_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	acpi_debugprint("acpi regs addr=0x%p\n", softstate->acpi_regs);
	acpi_debugprint("acpi regs handle=0x%p\n",softstate->acpi_regs_handle);

	if(!validate_regs(softstate)){
		cmn_err(CE_WARN, "acpi: attach for instance %d: failed check",
			instance);
		ddi_soft_state_free(acpi_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}


	/* you will probably want this at some point or another */
	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);

	mutex_init(&softstate->wait_mutex, NULL, MUTEX_DRIVER, NULL);

	cv_init(&softstate->wait_cv, NULL, CV_DRIVER, NULL);
	softstate->done_mutexes=1;

	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/XXXXXX
	 *
	 * To get an automatic matching link in /dev/XXXX, you need to 
	 * add your own line to /etc/devlink.tab
	 */
	 
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}


/* loop, with sleep, for things like "has power button been pressed?
 * called from either open(), or read().
 * Potentialy, will loop infinitely.
 * or until user-process-signal interrupts us.
 *
 * Return 1 on power signal, 0 on signal-abort or something.
 */
static int acpi_pollforevent(struct acpi_softstate *softstate)
{
	int tmpint=0;
	clock_t timeout_point, retval, onesec;

	onesec= 1 * drv_usectohz(1000000);

	enable_sci(softstate);

	ddi_dev_nintrs(softstate->dip, &tmpint);
	acpi_debugprint("Note: nintrs=%d\n", tmpint);

	/* Polling for powerbutton status.
	 * Note that in theory, can also get powerbutton status, from
	 * pci_config 0xa0, bit 9
	 * Other things of note:
	 *  pciconfig 0xA4, bit 2: battery weak or missing
	 */



	while(1){
		uint16_t reg16;
		void * regaddr = (uint8_t*)softstate->acpi_regs;
		reg16=ddi_get16(softstate->acpi_regs_handle,
		                regaddr + PM1_STS);
		acpi_debugprint("PM1_STS=0x%x\n", reg16);
		/* note: PM1_STS is normally 0x411 */

		if(reg16 & PM1_STS_PRBTN_STS){
			acpi_debugprint("Hey! power button pressed?! Clearing bit...");
			reg16 = PM1_STS_PRBTN_STS;
			ddi_put16(softstate->acpi_regs_handle,
		                regaddr + PM1_STS, reg16);

			
			acpi_debugprint("pollforevent returning 1");
			return 1;
			
		}

#ifdef DOESNTWORK
		/* err... we SHOULD be able to do this. but cant */
		if(reg16 && PM1_STS_PRBTNOR){
			printf("powerbutton override flagged. resetting");
			reg16=PM1_STS_PRBTNOR;
			ddi_put16(softstate->acpi_regs_handle,
		                regaddr + PM1_STS, reg16);
		}
#endif /* DOESNTWORK */

		reg16=ddi_get16(softstate->acpi_regs_handle,
		                regaddr + PM1_EN);
		acpi_debugprint("PM1_EN=0x%x\n", reg16);

		mutex_enter(&softstate->wait_mutex);

		timeout_point=ddi_get_lbolt();
		timeout_point += onesec;
		
		/* We EXPECT this to time out.
		 * We use this instead of delay(), because its the
		 * only way I know to allow for signals to interrupt.
		 */
		retval= cv_timedwait_sig(&softstate->wait_cv,
		                          &softstate->wait_mutex,
		                          timeout_point);
		mutex_exit(&softstate->wait_mutex);
		if(retval==0){
			acpi_debugprint("user signal detected\n");
			return 0;
			
		}
	}

	acpi_debugprint("pollforevent returning 0");
	return 0;
	
}

static int acpi_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	struct acpi_softstate *softstate;

   	acpi_debugprint("acpi: open called\n");

	/* even if you dont USE this, it is a check to see if we need
	 * to return ENXIO
	 */
	softstate=ddi_get_soft_state(acpi_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is technically NOT an error; it is expected.
		 * We return ENXIO to support "deferred attach"
		 */
		return (ENXIO);
	}

	/* use this HERE instead of in attach, when debugging new stuff*/
	/* validate_regs(softstate); */

	/* since I dont know of any ddi compliant way to have
	 * a "background thread" polling the hardware... and since
	 * drv_usecwait is a BUSYwait... and since
	 * putting a soft interrupt in a permenant loop is a no-no...
	 * AND since delay() is a  >>user context only<< call...
	 * Guess we'll have to be nasty, and have a demon just hold open
	 * the device, to activate power-switch polling.
	 * Maybe it should be in read(), though...?
	 */

	/*  enable_apic(softstate);  */

	/* return 0 on okay, or an errno otherwise */
   	return 0;
}



/* Normally, the only thing that should open and read the device,
 * is my acpi demon, "acpid".
 * It tries for a read of a single char.
 * We then loop forever, or until a powerbutton event is triggered,
 * at which time we send back an 'S' for Shutdown.
 *
 */
static int acpi_read(dev_t dev, struct uio* uiop, cred_t * credp){
	struct acpi_softstate *softstate;
	int instance=DEVT_TO_INST(dev);
	char *SHUTDOWN="S";
	int copystat;
	
	softstate=ddi_get_soft_state(acpi_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"read: cannot get softstate\n");
		return (EFAULT);
	}

	acpi_debugprint("acpi: read called\n");

	/* It is assumed that you are reading, to get ACPI events.
	* However, ACPI events are only enabled if we call enable_sci(),
	* in acpi_pollforevent(). BUT, it is only safe to do THAT,
	* if user is not using a laptop right now.
	* So, if feature not set... we cant do anything here.
	*/
	if(softstate->enable_powerbutton==0){
		return (ENXIO);
	}
	

	switch(acpi_pollforevent(softstate)){
		case 1:
			copystat=uiomove(SHUTDOWN, 1, UIO_READ, uiop);
			if(copystat !=0){
				acpi_debugprint("read: uiomove failed\n");
				return EFAULT;
			}
			acpi_debugprint("read: have sent SHUTDOWN msg\n");
			return 0;
		default:
			acpi_debugprint("read() will return 0 to user now\n");
			return 0;
	}

}

static int acpi_write(dev_t dev, struct uio *uiop, cred_t *credp){
/*	int instance=DEVT_TO_INST(dev); */

   	acpi_debugprint("acpi: write called\n");
	/* If you are handling things directly here rather than through
	 * physio, make sure that uiop->resid is set to 0 on good write
	 */
   	return (ENXIO);
}

static int acpi_close(dev_t dev, int openflags, int otyp, cred_t *credp){
   	acpi_debugprint("acpi: close called\n");
   	return (ENXIO);
}

/* This routine should never return :-> */
static void power_off(struct acpi_softstate *softstate)
{
	uint32_t reg32;
	void * regaddr;

	if(softstate->acpi_version==2){
		reg32=SLP_EN | SLP_TYPE_OFF;
	} else {
		reg32=SLP_EN | V1_SLP_TYPE_OFF;
	}
	reg32 |= 0x02;  /* tell BIOS to pay attention */

	regaddr=softstate->acpi_regs;
	regaddr += PM1_CNT;

	ddi_put32(softstate->acpi_regs_handle, regaddr, reg32);

	/* we should be off now :-) */
	
}


/* I'm not sure what happens with this.
 * It doenst actually seem to work. but SOME DAY.. it hopefully will.
 */
static void suspend_to_disk(struct acpi_softstate *softstate)
{
	uint32_t reg32;
	void * regaddr;

	if(softstate->acpi_version==2){
		reg32=SLP_EN | SLP_TYPE_DSUSP;
	} else {
		cmn_err(CE_WARN,"suspend_to_disk requires acpi v2\n");
		return;
	}
	reg32 |= 0x02;  /* tell BIOS to pay attention */

	regaddr=softstate->acpi_regs;
	regaddr += PM1_CNT;

	ddi_put32(softstate->acpi_regs_handle, regaddr, reg32);

	
}

static int acpi_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
                      cred_t *credp, int *rvalp){
	int instance=DEVT_TO_INST(dev);
	struct acpi_softstate *softstate;

	softstate=ddi_get_soft_state(acpi_glob_soft_handle,instance);

   	acpi_debugprint("acpi: ioctl called\n");

	if(cmd==1){
		power_off(softstate);
		/* does not return */
	}
	if(cmd==2){
		printf("DEBUG: suspend_to_disk will be attempted\n");
		delay(drv_usectohz(1000000));

		suspend_to_disk(softstate);
		/* does not return */
	}
	/* In theory, could have cmd==1 for hibernate.
	 * But need to check that it can even work for solaris
	 * Would be pretty cool for laptops.
	 * But would have to somehow make sure time got resynced later...?
	 */

	acpi_debugprint("ioctl called with cmd==%d\n");

   	return (ENXIO);
}

static int acpi_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct acpi_softstate *softstate;
   	instance=ddi_get_instance(dip);

	acpi_debugprint("acpi: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(acpi_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_NOTE,"acpi_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	acpi_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}

